@@ -1,4 +1,6 @@
-# User function -----------------------------------------------------------
+# Constructor -------------------------------------------------------------
+
+## External ---------------------------------------------------------------
 
 #' Extended wrapped facets
 #'
@@ -6,49 +8,105 @@
 #' extra options on axis drawing when scales are fixed.
 #'
 #' @inheritParams ggplot2::facet_wrap
-#' @param axes A \code{character} where axes should be drawn. Either
-#'   \code{"margins"} (default), \code{"rows"}, \code{"cols"} or \code{"full"}.
-#'   Only applies when the scale is free through the \code{scales} argument.
-#' @param remove_labels A \code{character} denoting what labels should be
-#'   removed when axes are repeated and redundant. Either \code{"none"}
-#'   (default), \code{"rows"}, \code{"cols"} or \code{"all"}. Only applies to
-#'   relevant position guides included with the \code{axes} argument when scales
-#'   are fixed.
+#' @param scales A `character(1)` or `logical(1)` whether scales are shared
+#'   across facets or allowed to vary. One of the following:
+#'   \describe{
+#'     \item{`"fixed"` or `FALSE`}{Scales are shared across all facets
+#'     (default).}
+#'     \item{`"free_x"`}{x-scales are allowed to vary.}
+#'     \item{`"free_y"`}{y-scales are allowed to vary.}
+#'     \item{`"free"` or `TRUE`}{Both scales can vary}
+#'   }
+#' @param axes A `character(1)` or `logical(1)` where axes should be drawn. One
+#'   of the following:
+#'   \describe{
+#'     \item{`"margins"` or `FALSE`}{Only draw axes at the outer margins
+#'       (default).}
+#'     \item{`"x"`}{Draw axes at the outer margins and all inner x-axes too.}
+#'     \item{`"y"`}{Draw axes at the outer margins and all inner y-axes too.}
+#'     \item{`"all"` or `TRUE`}{Draw the axes for every panel.}
+#'   }
+#' @param remove_labels A `character(1)` or `logical(1)` determining whether
+#'   axis text is displayed at inner panels. One of the following:
+#'   \describe{
+#'     \item{`"none"` or `FALSE`}{Display axis text at all axes (default).}
+#'     \item{`"x"`}{Display axis text at outer margins and all inner y-axes.}
+#'     \item{`"y"`}{Display axis text at outer margins and all inner x-axes.}
+#'     \item{`"all"` or `TRUE`}{Only display axis text at the outer margins.}
+#'   }
+#' @param trim_blank A `logical(1)`. When `TRUE` (default), does not draw rows
+#'   and columns containing no panels. When `FALSE`, the `nrow` and `ncol`
+#'   arguments are taken literally, even when there are more than needed to
+#'   fit all panels.
+#' @param strip An object created by a call to a strip function, such as
+#'   [`strip_vanilla`][strip_vanilla()].
 #'
 #' @return A \code{Facet} ggproto object that can be added to a plot.
 #' @export
 #' @family facetting functions
+#' @md
 #'
 #' @examples
 #' p <- ggplot(mpg, aes(displ, hwy)) + geom_point()
 #'
 #' # Repeat all axes for every facet
-#' p + facet_wrap2(vars(class), axes = "full")
+#' p + facet_wrap2(vars(class), axes = "all")
 #'
 #' # Repeat only y-axes
-#' p + facet_wrap2(vars(class), axes = "cols")
+#' p + facet_wrap2(vars(class), axes = "y")
 #'
 #' # Repeat axes without labels
-#' p + facet_wrap2(vars(class), axes = "full", remove_labels = "all")
+#' p + facet_wrap2(vars(class), axes = "all", remove_labels = "all")
 #'
 #' # Repeat axes without x-axis labels
-#' p + facet_wrap2(vars(class), axes = "full", remove_labels = "rows")
+#' p + facet_wrap2(vars(class), axes = "all", remove_labels = "x")
 facet_wrap2 <- function(
   facets, nrow = NULL, ncol = NULL,
   scales = "fixed", axes = "margins",
   remove_labels = "none",
   shrink = TRUE, labeller = "label_value",
   as.table = TRUE, drop = TRUE,
-  dir = "h", strip.position = "top"
+  dir = "h", strip.position = "top",
+  trim_blank = TRUE,
+  strip = strip_vanilla()
 ) {
+  new_wrap_facets(
+    facets, nrow, ncol,
+    scales, axes, remove_labels,
+    shrink, labeller,
+    as.table, drop, dir,
+    strip.position, strip,
+    trim_blank,
+    super = FacetWrap2
+  )
+}
 
-  strip.position <- match.arg(strip.position,
-                              c("top", "bottom", "left",  "right"))
+## Internal ---------------------------------------------------------------
+
+new_wrap_facets <- function(
+  facets, nrow, ncol,
+  scales, axes, rmlab,
+  shrink, labeller,
+  as.table, drop, dir,
+  strip.position, strip,
+  trim_blank,
+  params = list(), super = FacetWrap2
+) {
+  # Check arguments
   labeller <- .int$check_labeller(labeller)
-  facets <- .int$wrap_as_facets_list(facets)
+  strip.position <- arg_match0(
+    strip.position, c("top", "bottom", "left",  "right")
+  )
+  dir <- arg_match0(dir, c("h", "v"))
+  free  <- .match_facet_arg(scales, c("fixed", "free_x", "free_y", "free"))
+  axes  <- .match_facet_arg(axes,   c("margins", "x", "y", "all"))
+  rmlab <- .match_facet_arg(rmlab,  c("none", "x", "y", "all"))
+  strip <- assert_strip(strip)
 
-  # Take care of direction
-  dir <- match.arg(dir, c("h", "v"))
+  # Setup facet variables
+  facets   <- .int$wrap_as_facets_list(facets)
+
+  # Setup dimensions
   if (identical(dir, "v")) {
     nrow_swap <- ncol
     ncol_swap <- nrow
@@ -58,43 +116,22 @@ facet_wrap2 <- function(
     nrow <- .int$sanitise_dim(nrow)
     ncol <- .int$sanitise_dim(ncol)
   }
+  dim <- if (trim_blank) NULL else c(nrow %||% NA, ncol %||% NA)
 
-  # Setup free scales
-  scales <- match.arg(scales, c("fixed", "free_x", "free_y", "free"))
-  free <- list(
-    x = any(scales %in% c("free_x", "free")),
-    y = any(scales %in% c("free_y", "free"))
-  )
-
-  # Setup axis positions
-  axes <- match.arg(axes, c("margins", "rows", "cols", "full"))
-  axes <- list(
-    x = any(axes %in% c("rows", "full")),
-    y = any(axes %in% c("cols", "full"))
-  )
-
-  rmlab <- match.arg(remove_labels, c("none", "rows", "cols", "all"))
-  rmlab <- list(
-    x = any(rmlab %in% c("rows", "all")),
-    y = any(rmlab %in% c("cols", "all"))
-  )
+  # Make list of parameters
+  params <- c(params, list(
+    facets = facets, free = free, as.table = as.table,
+    strip.position = strip.position, drop = drop,
+    ncol = ncol, nrow = nrow, dim = dim,
+    labeller = labeller, dir = dir,
+    axes = axes, rmlab = rmlab
+  ))
 
   ggproto(
-    NULL, FacetWrap2,
+    NULL, super,
     shrink = shrink,
-    params = list(
-      facets = facets,
-      free = free,
-      as.table = as.table,
-      strip.position = strip.position,
-      drop = drop,
-      ncol = ncol,
-      nrow = nrow,
-      labeller = labeller,
-      dir = dir,
-      axes = axes,
-      rmlab = rmlab
-    )
+    strip = strip,
+    params = params
   )
 }
 
@@ -110,20 +147,6 @@ facet_wrap2 <- function(
 #' @rdname ggh4x_extensions
 FacetWrap2 <- ggproto(
   "FacetWrap2", FacetWrap,
-  setup_strips = function(layout, params, theme) {
-    # Format labels and render strips
-    if (length(params$facets) == 0) {
-      labels_df <- .int$new_data_frame(list("(all)" = "(all)"), n = 1)
-    } else {
-      labels_df <- layout[names(params$facets)]
-    }
-    attr(labels_df, "facet") <- "wrap"
-    render_strips(
-      structure(labels_df, type = "rows"),
-      structure(labels_df, type = "cols"),
-      params$labeller, theme
-    )
-  },
   setup_aspect_ratio = function(coord, free, theme, ranges) {
     # Rules for aspect ratio
     aspect_ratio <- theme$aspect.ratio
@@ -157,48 +180,52 @@ FacetWrap2 <- ggproto(
 
     layout
   },
-  setup_panel_table = function(self, panels, empty, position,
-                               theme, coord, ranges, free) {
-    # Puts panels into a gtable with names and spacing
-    dim <- dim(empty)
-    panel_table <- empty
-    panel_table[position] <- panels
+  setup_panel_table = function(self, panels, layout, #empty, position,
+                               theme, coord, ranges, params) {
+    ncol <- params$dim[2] %||% max(c(layout$.LEFT, layout$.RIGHT))
+    nrow <- params$dim[1] %||% max(c(layout$.TOP,  layout$.BOTTOM))
 
-    aspect_ratio <- self$setup_aspect_ratio(coord, free, theme, ranges)
+    aspect <- self$setup_aspect_ratio(coord, params$free, theme, ranges)
 
-    empties <- apply(panel_table, c(1, 2), function(x){.int$is.zero(x[[1]])})
-    panel_table <- gtable_matrix(
-      "layout", panel_table,
-      widths  = unit(rep(1, dim[2]), "null"),
-      heights = unit(rep(abs(aspect_ratio), dim[1]), "null"),
-      respect = attr(aspect_ratio, "respect"), clip = coord$clip,
-      z = matrix(1, ncol = dim[2], nrow = dim[1])
+    respect <- params$respect %||% attr(aspect, "respect") %||% FALSE
+    widths  <- params$widths  %||% unit(1, "null")
+    heights <- params$heights %||% unit(abs(aspect %||% 1), "null")
+
+    panel_table <- gtable(
+      widths  = rep_len(widths, ncol),
+      heights = rep_len(heights, nrow),
+      respect = respect
+    )
+    panel_table <- gtable_add_grob(
+      panel_table, panels,
+      t = layout$.TOP,  b = layout$.BOTTOM,
+      l = layout$.LEFT, r = layout$.RIGHT,
+      z = 1, clip = coord$clip, name = paste0("panel-", seq_along(panels))
     )
-    panel_table$layout$name <- paste0("panel-", rep(seq_len(dim[2]), dim[1]),
-                                      "-", rep(seq_len(dim[1]), each = dim[2]))
-
     panel_table <- gtable_add_col_space(
       panel_table, calc_element("panel.spacing.x", theme)
     )
     panel_table <- gtable_add_row_space(
       panel_table, calc_element("panel.spacing.y", theme)
     )
-    list(panels = panel_table, empty = empties)
+    panel_table
   },
-  setup_axes = function(axes, empty, position, layout,
-                        params, empties, theme) {
+  setup_axes = function(axes, layout, params, theme) {
+
     # Takes rendered axes and picks the right ones for the facets
-    dim <- dim(empty)
-    nrow <- dim[1]
-    ncol <- dim[2]
+    nrow  <- max(layout$ROW)
+    ncol  <- max(layout$COL)
+    index <- cbind(layout$ROW, layout$COL)
+    empties <- matrix(TRUE, nrow, ncol)
+    empties[index] <- FALSE
 
     # Initialise empty axes
-    top <- bottom <- left <- right <- empty
+    top <- bottom <- left <- right <- matrix(list(zeroGrob()), nrow, ncol)
     # Fill axes by scale ID
-    top[position]    <- axes$x$top[layout$SCALE_X]
-    bottom[position] <- axes$x$bottom[layout$SCALE_X]
-    left[position]   <- axes$y$left[layout$SCALE_Y]
-    right[position]  <- axes$y$right[layout$SCALE_Y]
+    top[index]    <- axes$x$top[layout$SCALE_X]
+    bottom[index] <- axes$x$bottom[layout$SCALE_X]
+    left[index]   <- axes$y$left[layout$SCALE_Y]
+    right[index]  <- axes$y$right[layout$SCALE_Y]
 
     repeat_x <- params$free$x | params$axes$x
     repeat_y <- params$free$y | params$axes$y
@@ -223,9 +250,15 @@ FacetWrap2 <- ggproto(
       right[, -ncol] <- lapply(right[, -ncol], purge_guide_labels)
     }
 
+    # Measure after deletion of axes, but before placing them back.
+    # This assures that the gap between dangling panels isn't too big.
+    measurements <- .measure_axes(list(top = top, bottom = bottom,
+                                       left = left, right = right))
+
     # Place back necessary axes that were removed prior, e.g. when there are
     # missing panels so the marginal axes should not be the most extreme one.
     if (any(empties)) {
+
       # Find first non-empty panel
       first_row <- which(apply(empties, 1, any))[1] - 1
       first_col <- which(apply(empties, 2, any))[1] - 1
@@ -239,8 +272,8 @@ FacetWrap2 <- ggproto(
                                  nrow)
 
       # Find corresponding axes
-      row_axes <- axes$x$bottom[layout$SCALE_X[row_panels]]
-      col_axes <- axes$y$right[layout$SCALE_Y[col_panels]]
+      row_axes <- axes$x$bottom[layout$SCALE_X[first_row]]
+      col_axes <- axes$y$right[layout$SCALE_Y[first_col]]
 
       inside <- (theme$strip.placement %||% "inside") == "inside"
 
@@ -269,24 +302,10 @@ FacetWrap2 <- ggproto(
       top = top,
       bottom = bottom,
       left = left,
-      right = right
+      right = right,
+      measurements = measurements
     )
   },
-  measure_axes = function(axes) {
-    top <- unit(
-      apply(axes$top,    1, max_height, value_only = TRUE), "cm"
-    )
-    bottom <- unit(
-      apply(axes$bottom, 1, max_height, value_only = TRUE), "cm"
-    )
-    left <- unit(
-      apply(axes$left,   2, max_width,  value_only = TRUE), "cm"
-    )
-    right <- unit(
-      apply(axes$right,  2, max_width,  value_only = TRUE), "cm"
-    )
-    list(top = top, bottom = bottom, left = left, right = right)
-  },
   attach_axes = function(panel_table, axes, sizes) {
     panel_table <- .int$weave_tables_row(
       panel_table, axes$top,   -1, sizes$top,    "axis-t", 3
@@ -302,74 +321,6 @@ FacetWrap2 <- ggproto(
     )
     panel_table
   },
-  attach_strips = function(panel_table, strips, position, empty, params, theme,
-                           clip = "off", sizes) {
-    # Setup parameters
-    strip_padding <- convertUnit(theme$strip.switch.pad.wrap, "cm")
-    strip_name <- paste0("strip-", substr(params$strip.position, 1, 1))
-    strip_mat <- empty
-    strip_pos <- params$strip.position
-    strip_mat[position] <- unlist(unname(strips),
-                                  recursive = FALSE)[[strip_pos]]
-
-    if (strip_pos %in% c("top", "bottom")) {
-      inside_x <- (calc_element("strip.placement.x", theme) %||% "inside")
-      inside_x <- inside_x  == "inside"
-      if (strip_pos == "top") {
-        offset  <- if (inside_x) -1 else -2
-        padding <- sizes$top
-      } else {
-        offset <- if (inside_x) 0 else 1
-        padding <- sizes$bottom
-      }
-      strip_height <- unit(
-        apply(strip_mat, 1, max_height, value_only = TRUE), "cm"
-      )
-
-      # Add top/bottom strips
-      panel_table <- .int$weave_tables_row(
-        panel_table, strip_mat, offset, strip_height, strip_name, 2, clip
-      )
-
-      if (!inside_x) {
-        # Add padding
-        padding[as.numeric(padding) != 0] <- strip_padding
-        panel_table <- .int$weave_tables_row(
-          panel_table, row_shift = offset, row_height = padding
-        )
-      }
-
-    } else {
-      inside_y <- (calc_element("strip.placement.y", theme) %||% "inside")
-      inside_y <- inside_y == "inside"
-      if (strip_pos == "left") {
-        offset <- if (inside_y) -1 else -2
-        padding <- sizes$left
-      } else {
-        offset <- if (inside_y) 0 else 1
-        padding <- sizes$right
-      }
-      padding[as.numeric(padding) != 0] <- strip_padding
-      strip_width <- unit(
-        apply(strip_mat, 2, max_width, value_only = TRUE), "cm"
-      )
-
-      # Add left/right strips
-      panel_table <- .int$weave_tables_col(
-        panel_table, strip_mat, offset, strip_width, strip_name, 2, clip
-      )
-
-      if (!inside_y) {
-        padding[as.numeric(padding) != 0] <- strip_padding
-        panel_table <- .int$weave_tables_col(
-          panel_table, col_shift = offset, col_width = padding
-        )
-      }
-    }
-
-    panel_table
-
-  },
   finish_panels = function(self, panels, layout, params, theme) {
     panels
   },
@@ -380,38 +331,37 @@ FacetWrap2 <- ggproto(
       stop(.int$snake_class(coord), " doesn't support free scales",
            call. = FALSE)
     }
-    # browser()
-
+    strip  <- self$strip
     layout <- self$setup_layout(layout, coord, params)
 
     # Setup parameters
     ncol <- max(layout$COL)
     nrow <- max(layout$ROW)
-    n <- nrow(layout)
-    panel_order <- order(layout$ROW, layout$COL)
-    layout <- layout[panel_order, ]
-    panels <- panels[panel_order]
-    panel_pos <- .int$convertInd(layout$ROW, layout$COL, nrow)
+    layout$.TOP  <- layout$.BOTTOM <- layout$ROW
+    layout$.LEFT <- layout$.RIGHT  <- layout$COL
+    if (!is.null(params$dim)) {
+      if (is.na(params$dim[1])) {
+        params$dim[1] <- nrow
+      }
+      if (is.na(params$dim[2])) {
+        params$dim[2] <- ncol
+      }
+    }
 
     # Setup panels
-    empty_table <- matrix(list(zeroGrob()), nrow = nrow, ncol = ncol)
-    panel_table <- self$setup_panel_table(panels, empty_table, panel_pos,
-                                          theme, coord, ranges, params$free)
-    empties <- panel_table$empty
-    panel_table <- panel_table$panels
+    panel_table <- self$setup_panel_table(
+      panels, layout, theme, coord, ranges, params
+    )
 
     # Deal with axes
     axes <- render_axes(ranges, ranges, coord, theme, transpose = TRUE)
-    axes <- self$setup_axes(axes, empty_table, panel_pos, layout,
-                            params, empties, theme)
-    sizes <- self$measure_axes(axes)
-    panel_table <- self$attach_axes(panel_table, axes, sizes)
+    axes <- self$setup_axes(axes, layout, params,theme)
+    panel_table <- self$attach_axes(panel_table, axes, axes$measurements)
 
     # Deal with strips
-    strips <- self$setup_strips(layout, params, theme)
-    panel_table <- self$attach_strips(
-      panel_table, strips, panel_pos, empty_table, params, theme,
-      clip = coord$clip, sizes
+    strip$setup(layout, params, theme, type = "wrap")
+    panel_table <- strip$incorporate_wrap(
+      panel_table, params$strip.position, clip = coord$clip, axes$measurements
     )
 
     self$finish_panels(panels = panel_table, layout = layout,
@@ -441,3 +391,28 @@ purge_guide_labels <- function(guide) {
   guide$height <- guide$vp$height <- sum(axis$heights)
   guide
 }
+
+.match_facet_arg <- function(value, options, x = 2, y = 3, both = 4, neither = 1,
+                             nm = deparse(substitute(value))) {
+  if (is.logical(value) && length(value) == 1 && !is.na(value)) {
+    if (value) {
+      value <- options[both]
+    } else {
+      value <- options[neither]
+    }
+  } else {
+    value <- rlang::arg_match0(value, options, arg_nm = nm)
+  }
+  list(
+    x = any(value %in% options[c(x, both)]),
+    y = any(value %in% options[c(y, both)])
+  )
+}
+
+.measure_axes <- function(axes) {
+  top    <- unit(apply(axes$top,    1, max_height, value_only = TRUE), "cm")
+  bottom <- unit(apply(axes$bottom, 1, max_height, value_only = TRUE), "cm")
+  left   <- unit(apply(axes$left,   2, max_width,  value_only = TRUE), "cm")
+  right  <- unit(apply(axes$right,  2, max_width,  value_only = TRUE), "cm")
+  list(top = top, bottom = bottom, left = left, right = right)
+}
